insert into dbversion(version, release, description)
values (10, '2016-02-22 15:56:28.358587+00', 'Work In Progress');

insert into origin(url, hash)
values ('https://forge.softwareheritage.org', hash_sha1('https://forge.softwareheritage.org'));
